/* The copyright in this software is being made available under the BSD
* License, included below. This software may be subject to other third party
* and contributor rights, including patent rights, and no such rights are
* granted under this license.
*
* <OWNER> = Apple Inc.
* <ORGANIZATION> = Apple Inc.
* <YEAR> = 2014
*
* Copyright (c) 2014, Apple Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  * Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*  * Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*  * Neither the name of the <ORGANIZATION> nor the names of its contributors may
*    be used to endorse or promote products derived from this software without
*    specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
* THE POSSIBILITY OF SUCH DAMAGE.
*/

/*!
*************************************************************************************
* \file Reshaper.H
*
* \brief
*    Class Header for Reshaper
*
* \author
*     - Alexis Michael Tourapis         <atourapis@apple.com>
*
*************************************************************************************
*/


#ifndef __Reshaper_H__
#define __Reshaper_H__

#include <vector>
#include <sstream>
#include "Global.H"
#include "Parameters.H"
#include "Frame.H"
#if TCH_CROSSPLANE_CRCB
#include "CrossPlaneLut.H"
#endif

#if DI_RESHAPER_m37269_m37267
using namespace std;

typedef struct _unified_luma_reshape_info
{
  int POC;
  int reshapePresentFlag;
  int reshape_num_ranges_minus1;
  int reshape_global_offset_val;
  int reshape_range_val[MAX_N_SEGMENTS];
  int coeff_log2_offset_minus2;  // TBD this should range to minus 1 to allow a value of 1 bit for PL
  int reshape_continuity_flag;

  int constantCoeffSignalledInt[MAX_N_SEGMENTS+1];
  int constantCoeffSignalledFrac[MAX_N_SEGMENTS+1];
  int linearCoeffSignalledInt;
  int linearCoeffSignalledFrac;
  // derived values
  int fractionBits[3];
  int constantCoeffSignalled[MAX_N_SEGMENTS+1];
  int linearCoeffSignalled;
  int nAllSegmentLengths;
  int allSegmentLengths[MAX_N_SEGMENTS];
  long polyCoeff[MAX_N_SEGMENTS][MAX_PRED_ORDER+1];

  bool firstSegmentFlat;
  bool lastSegmentFlat;
  
  int luma_bit_depth;  
  int reshape_equal_ranges;
  int inverseLumaReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE];
  int inverseChromaReshapingLUT[2][MAX_LUMA_RESHAPING_LUT_SIZE];
} UnifiedLumaReshapeInfo;

#if ARRS_MODEL_OPTIMIZATION
typedef  struct _reshaperParams
  {
    float min[3];
    float max[3];
    float temporalPeakBrightness;
    float temporalBlackLevel;
    float temporalAvgBrightness;
    double temporalGamma;
    double reshaperAlpha;
    bool  newScene;
  } reshaperParams;
#endif


typedef struct _reshapingInfo {
  int     methods;                  //! reshaping method: 0: disable, 1: enable
  int     m_reshapeMode;
  char	  metaFile[MAX_LINE_LEN];   //! scene info related metaFile  
#if MODE0_UPDATE
  int		m_intraPeriod;			//! intra period based on frame rate
  int		m_srcPrimary;			//! original src primary 
#endif
#if QC_CBCR_m37064 
  char          atfParamsLumaFilename[MAX_LINE_LEN];
  char          atfParamCbCrFilename[MAX_LINE_LEN];
  bool			m_isChromaHistScale;		   //! The value to scale chroma histogram (Single value for U and V for now)
  float			m_chromaHistScaleCb;		   //! The value to scale chroma histogram for U)
  float			m_chromaHistScaleCr;		   //! The value to scale chroma histogram for V)
  float			m_chromaHistOffsetCb;		   //! The offset value for U					 
  float			m_chromaHistOffsetCr;          //! The offset value for V
#endif
#if ARRS_MODEL_OPTIMIZATION
  float     blackLevelLiftTarget; //! black level lift of the target display (e.g. compare to BT.1886)
  float     gammaTarget;          //! gamma value of the target display (e.g. compare to BT.1886)
  float     blackLevelLiftInput;  //! equivalent of the black level lift for an estimated gamma function of the input TF (e.g. compare to BT.2020 TF)
  float     alphaInput;           //! equivalent of the alpha value for a power function estimation of the input TF (e.g. compare to BT.2020 TF)
#endif
}ReshapingInfo;


typedef struct _sceneBlkData {
	int blockSize;
	int blockWidth;
	int blockHeight;
	double *blockDiff;
	double *blockStd;
	double *blockMean;
#if ARRS_MODEL_OPTIMIZATION
  double *blockMinIntensity;
  double *blockMaxIntensity;
  double *blockAvgIntensity;
#endif
}SceneBlkData;

typedef struct _sceneData {
#if MODE0_UPDATE
	double	prev_org_I_CH;
	double	curr_org_I_CH;
	int		prev_frameNo;
#endif
  int     sceneFirstFno;            //! scene first frame no.
  int     srcPrimary;               //! source primary: 0-709, 1-BT.2020, 3-P3
  int	  sceneIsFades;				//! detect fading frames
}SceneData;

// ====================================================================================================================
// Class definition
// ====================================================================================================================
/* class for creating some temporary matrices to be used by function MinimalMSE_Fit() in class TEncReshape. */
class TempMatrices
{
public:
  /* members */
  int m_nSegments;
  int m_nParams;
  int m_nSamples;

  double**		Param2A;
  double**		Param2B;
  double**		Param2C;
  double**		Param2IntermediateB;
  double**		IntermediateB2C;
  double**		Param2StackedABC;
  double**		StackedABC2InterleavedABC;
  double**		InterleavedABC2Y;
  double**		Param2Y;
  double**		Param2Y_t;
  double**		M;
  double**		M_t;
  double**		M_txM;
  double**		Param2Y_txParam2Y;
  double**        temp1;
  double*			SegmentLowerEndPoints;
  double*			Y1;
  double*			Y2;
  double*			M_txY2;
  double*			ParamOpt_temp;
  double*			ParamOpt;
  double*			Param2Y_txlut;
  double*			V;
  double*			V_t;
  double*			Y;
  double*			pred;
  double*         __x;
  double*         x;
  double*         error;
  double*         Param2YxParamOpt;
  double*         final_pred;
  /* methods */
  TempMatrices(int nSegments, int nParams, int nSamples);        // constructor
  ~TempMatrices();                                               // destructor
};


class Reshaper {
private:
  int  m_inverseLumaReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE];
#if ARRS_MODEL_OPTIMIZATION
  int  m_forwardLumaReshapingLUT[MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE];
  int m_offsetArray[4];
#else
  int  m_forwardLumaReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE];
#endif
  int  m_inverseChromaReshapingLUT[2][MAX_LUMA_RESHAPING_LUT_SIZE];
#if ARRS_MODEL_OPTIMIZATION
  int  m_forwardChromaReshapingLUT[2][MAX_LUMA_FORWARD_RESHAPING_LUT_SIZE];
#else
  int  m_forwardChromaReshapingLUT[2][MAX_LUMA_RESHAPING_LUT_SIZE];
#endif
#if QC_CBCR_m37064
  float         m_chromaHistScaleCb;
  float         m_chromaHistScaleCbInv;
  float         m_chromaHistScaleCr;
  float         m_chromaHistScaleCrInv;
  float         m_chromaHistOffsetCb;
  float         m_chromaHistOffsetCr;
  bool		    m_flagScaleChromaHist;
#endif
#if TCH_CROSSPLANE_CRCB
  double  m_temporalBa;
  double  m_peakLum;
  double m_last_written_Ba;
  int m_reshaperRangeId;
#endif
#if ARRS_MODEL_OPTIMIZATION
  reshaperParams m_reshaperParams;
#endif
  /* some helper functions */
  double    minVec            (vector<double>&);
  double    maxVec            (vector<double>&);
  void      setNewUpperBound  (vector<double>& vec, double upBnd);
  void      setNewLowerBound  (vector<double>& vec, double loBnd);
  double    sumVec            (vector<double>, int, int);
  void      mulMatMat         (double** a, double** b, double** c, int a_rows, int b_cols, int a_cols);  // Mat x Mat
  void      mulMatVec         (double** a, double* b, double* c, int a_rows, int a_cols);               // Mat x Vec
  void      transpose         (double**, double**, int, int);
  double    dotProd           (double*, double*, int);
  void      LUFactorization   (double** A_txA, double* A_txY, double* X, int N);
  
  /* find optimum params */
  void MinimalMSE_Fit (double* lut, int lut_len,vector<double> SegmentLengths, bool smoothness_flag,bool force_initial_value_flag, int nFractionBitsA, int  nFractionBitsB, TempMatrices* cTempMatrices, vector<double>&  A0S,double& B0,	double& mse	);
  void blockproc(const double *input, int inp_wid, int inp_hei, SceneBlkData sceneBLK);
#if MODE0_UPDATE
  void frameAnalysis(const Frame *inp);
#endif
  void sceneAnalysis(const Frame *inp, double *org_CL, double *org_CH, double *scale_factor, double *power_alpha, SceneData metaData);
#if ARRS_MODEL_OPTIMIZATION
  void getReshaperParameters(int frameNumber, float *Ycomp, int width, int height, ReshapingInfo *reshaping);
  void generateLUTs (int frameNumber, float *preScalingMinMax, float *postScalingMinMax, int *prebitDepth, int *postbitDepth, SampleRange preSampleRange, SampleRange postSampleRange, ReshapingInfo reshaping );
  void blockproc(float *input, int inp_wid, int inp_hei, SceneBlkData sceneBLK, double TFGamma);
  void floatToFixed( float *input, void *output, int *bitDepth, SampleRange sampleRange, float *minMax, int *size, int comp );
  void fixedToFloat( void *input, float *output, int *bitDepth, SampleRange sampleRange, float *minMax, int *size, int comp );
#endif
public:
  int  m_numOfScenes;
  int  m_sceneNum;
#if MODE0_UPDATE
  SceneData	  m_sceneData;
#else
  SceneData  *m_sceneData;  //store all scene info. in a lip
#endif
  UnifiedLumaReshapeInfo *m_reshapeData; //store all reshape info. (scene based)
  int m_startFrame;         //handle start frame not zero, note: output metadata: POC always starts from zero to be aligned with POC at encoder.
#if TCH_CROSSPLANE_CRCB
  int m_recomputeReshapeParams;
#endif
  Reshaper();
  ~Reshaper();
  void AutomaticMinimalMSE_Fitting(int full_lut_len, double* target_lut, int input_bit_depth, int nFractionBits, bool smoothness_flag, UnifiedLumaReshapeInfo* pUnifiedLumaReshapeInfo);
  void WriteMetaData(char *filename, int POC, UnifiedLumaReshapeInfo *pUnifiedLumaReshape);

  
  void DetermineAllFractionBits(int coeff_log2_offset_minus2,int bitDepth,int fractionBits[MAX_PRED_ORDER+1]);
  void DeriveAllSegments(int nSegmentSignaled,int SegmentLengths[MAX_N_SEGMENTS],int global_offset_val,int total_length,int *pNumberAllSegments,int pAllSegmentLengths[MAX_N_SEGMENTS],bool *pFirstSegmentFlat,bool *pLastSegmentFlat);
  void DeriveSmoothPolynomialCoefficients(int nSegmentSignaled, int SegmentLengths[MAX_N_SEGMENTS], int constantCoeff[1+MAX_N_SEGMENTS], int linearCoeff, bool FirstSegmentFlatFlag, bool LastSegmentFlatFlag,int nFractionBits[3], long polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER+1]);
  void DerivePiecewiseLinearPolynomialCoefficients(int nSegmentSignaled, int SegmentLengths[MAX_N_SEGMENTS], int constantCoeff[1+MAX_N_SEGMENTS],bool FirstSegmentFlatFlag,bool LastSegmentFlatFlag,int nFractionBits[3], long polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER+1]);

  void BuildInverseReshapingLUT(int nSegmentAll, int allSegmentLengths[MAX_N_SEGMENTS],int nFractionBits[3], long polyCoeff[MAX_LUMA_RESHAPE_SEGMENTS][MAX_PRED_ORDER+1],int lut_size,int InverseReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE]);
  void BuildForwardReshapingLUT(int InverseReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int ForwardReshapingLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size);
  void BuildIdentityLUT(int LUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size);

#if MODE0_UPDATE
  void InitSceneData();
#else
  void readSceneMetadata(char *metaFile);
#endif
  void forwardReshaping(const Frame *inp, int POC, ReshapingInfo reshaping);
  void process ( Frame* out, const Frame *inp, ReshapingInfo reshaping);

#if ARRS_MODEL_OPTIMIZATION
  void process ( Frame* out, Frame *inp, ReshapingInfo reshaping, int POC);
#endif

#if TCH_CROSSPLANE_CRCB
  void   BuildInverseYlogLUT(double ForwardYlogLUT[MAX_LUMA_RESHAPING_LUT_SIZE],double InverseYlogLUT[MAX_LUMA_RESHAPING_LUT_SIZE],int lut_size);
  float  computeGlogParameters(int frameNumber, const Frame &inp) ;
  double TchComputeMeanY(const uint16 * yin, int width, int height, int size, int frameNumber, SampleRange range); 
  void   setParamsTCHYATF(Tch::PWModel &pw);
#endif

#if QC_CBCR_m37064 //DI modified/moved QCOM implementations
  void  setChromaScaling(float *trg_scale, float *trg_offset, float src_scale, float src_offset);
  float evaluateFloatRepn(int x, int const numBitsVal, int const numBitsFrac);
  int   evaluateIntRepn(float const x, int const numBitsVal, int const numBitsFrac);
  void  writeCbCrATF(char *filename, int POC);
  void  setRangeId(int rangeId) { m_reshaperRangeId = rangeId; };
  int   getRangeId(void) { return m_reshaperRangeId; };
  void  setParamsATF(ReshapingInfo *reshape);
#endif

};
#if MODE0_UPDATE
struct fixedPrec{
    int val;
    int fracBitDepth;
    fixedPrec() : val(0), fracBitDepth(0) {}
    fixedPrec(int x, int y) : val(x), fracBitDepth(y) {}

    fixedPrec operator+(const fixedPrec &x) {
        assert( x.fracBitDepth == fracBitDepth);
        return fixedPrec(val + x.val, fracBitDepth);
    }
    fixedPrec operator-(const fixedPrec &x) {
        assert( x.fracBitDepth == fracBitDepth);
        return fixedPrec(val - x.val, fracBitDepth);
    }
    fixedPrec operator*(const fixedPrec &x) {
        return fixedPrec(val * x.val, fracBitDepth + x.fracBitDepth);
        return *this;
    }
    fixedPrec operator/(const fixedPrec &x) {
        return fixedPrec(val / x.val, fracBitDepth - x.fracBitDepth);
        return *this;
    }
    fixedPrec & operator=(const fixedPrec &x) { val = x.val; fracBitDepth = x.fracBitDepth; return *this;}

    bool operator<(const fixedPrec &x) { assert( fracBitDepth == x.fracBitDepth); return val < x.val; }
    bool operator<=(const fixedPrec &x) { assert( fracBitDepth == x.fracBitDepth); return val <= x.val; }
    bool operator>(const fixedPrec &x) { assert( fracBitDepth == x.fracBitDepth); return val > x.val; }
    bool operator>=(const fixedPrec &x) { assert( fracBitDepth == x.fracBitDepth); return val >= x.val; }
    bool operator==(const fixedPrec &x) { return (val == x.val) && (fracBitDepth == x.fracBitDepth) ; }

    static fixedPrec rshift(fixedPrec const var, int x) { return fixedPrec(var.val << x, var.fracBitDepth + x);}
};
#endif
#endif
#endif